require("util.nut");

class AirManager {
	cargo_passengers = null;
	airportTypes = [AIAirport.AT_INTERCON, AIAirport.AT_INTERNATIONAL, AIAirport.AT_METROPOLITAN,
						  AIAirport.AT_LARGE, AIAirport.AT_COMMUTER, AIAirport.AT_SMALL,
						  AIAirport.AT_HELISTATION, AIAirport.AT_HELIDEPOT, AIAirport.AT_HELIPORT
	                      ];
	smallAirportTypes = [AIAirport.AT_COMMUTER, AIAirport.AT_SMALL];
	helicopterAirportTypes = [AIAirport.AT_HELISTATION, AIAirport.AT_HELIDEPOT, AIAirport.AT_HELIPORT];
	
	static airportTypeMaxVehicles = {
		[AIAirport.AT_HELIPORT]=3,
		[AIAirport.AT_HELIDEPOT]=3,
		[AIAirport.AT_HELISTATION]=5,
		[AIAirport.AT_SMALL]=2,
		[AIAirport.AT_COMMUTER]=3,
		[AIAirport.AT_LARGE]=4,
		[AIAirport.AT_METROPOLITAN]=6,
		[AIAirport.AT_INTERNATIONAL]=8,
		[AIAirport.AT_INTERCON]=10
	}
	townBlacklist = [];
	airportUpgradeBlacklist = [];
	static function GetNumberOfPlanes(stationID);
	static function SupportsExtraPlanes(stationID);
	static function IsClearOrAirport(tile, station);
	function BuildAirRoute();
	function UpgradePlanes();
	function BuildNewAirport();
	function BuildAirport(townID);
	function UpgradeAirports();
	function ConnectAirports(air1, air2);
	function BuildAirportType(townID, type);
	
	constructor() {
		local cargos = AICargoList();
		for (local cargo = cargos.Begin(); !cargos.IsEnd(); cargo = cargos.Next()) {
			if (AICargo.HasCargoClass(cargo, AICargo.CC_PASSENGERS)) {
				if (cargo_passengers != null){
					AILog.Warning("multiple passenger cargos detected, airplanes will not work correctly");
				}
				cargo_passengers = cargo;
			}
		}
	}
}

function AirManager::GetNumberOfPlanes(stationID){
	local vehicles = AIVehicleList_Station(stationID);
	vehicles.Valuate(AIVehicle.GetVehicleType);
	vehicles.KeepValue(AIVehicle.VT_AIR);
	return vehicles.Count();
}

function AirManager::SupportsExtraPlanes(stationID){
	local numberOfPlanes = AirManager.GetNumberOfPlanes(stationID);
	local maxNumberOfPlanes = AirManager.airportTypeMaxVehicles[AIAirport.GetAirportType(AIBaseStation.GetLocation(stationID))];
	return numberOfPlanes < maxNumberOfPlanes;
}

function AirManager::IsClearOrAirport(tile, station){
	return (AITile.IsBuildable(tile) && AITile.GetSlope(tile)==AITile.SLOPE_FLAT) ||
	       (station != null && AIAirport.IsAirportTile(tile) && AIStation.GetStationID(tile) == station)
}

function AirManager::BuildAirRoute(){
	if (UpgradeAirports()){
		return true;
	}
	local stations = AIStationList(AIStation.STATION_AIRPORT);
	local numStations = stations.Count();
	stations.Valuate(Util.StationWaitingCargo);
	stations.KeepAboveValue(50);
	local stationsWithCargo = stations.Count();
	stations.Valuate(SupportsExtraPlanes);
	stations.KeepValue(1);
	local stations2 = AIStationList(AIStation.STATION_AIRPORT);
	stations2.Valuate(GetNumberOfPlanes);
	stations2.KeepValue(0);
	AILog.Info("numStations: "+numStations+" stations with cargo: "+stationsWithCargo + " stations without planes: "+stations2.Count());
	if (numStations>=3 && stationsWithCargo<=2 && stations2.Count()==0){
		//wait
		return false;
	}
	stations.AddList(stations2);
	if (stations.Count()>=2 && numStations >= 3){
		stations.Valuate(AIBase.RandItem); //sort randomly
		stations.Sort(AIList.SORT_BY_VALUE, true);
		local station1=stations.Begin();
		local station2=stations.Next();
		if (!ConnectAirports(AIBaseStation.GetLocation(station1), AIBaseStation.GetLocation(station2))){
			BuildNewAirport();
		}
	}
	else {
		AILog.Info("not enough stations: "+stations.Count()+"/"+numStations);
		return BuildNewAirport();
	}
}

function AirManager::UpgradePlanes(){
	local engines = AIEngineList(AIVehicle.VT_AIR);
	engines.Valuate(AIEngine.GetCapacity);
	engines.Sort(AIList.SORT_BY_VALUE, false);
	
	local planes = AIVehicleList();
	planes.Valuate(AIVehicle.GetVehicleType);
	planes.KeepValue(AIVehicle.VT_AIR);
	
	local airports = AIStationList(AIStation.STATION_AIRPORT);
	
	foreach (plane, _ in planes){
		if (Util.CapLength(AIVehicle.GetName(plane),6)=="delete"){
			continue;
		}
		foreach (en, _ in engines){
			if (AIVehicle.GetCapacity(plane, cargo_passengers)<AIEngine.GetCapacity(en) && Money.ReserveMoney(AIEngine.GetPrice(en)) &&
				AIEngine.GetPlaneType(AIVehicle.GetEngineType(plane))==AIEngine.GetPlaneType(en)){
				airports.Valuate(AIStation.GetDistanceManhattanToTile,AIVehicle.GetLocation(plane));
				airports.Sort(AIList.SORT_BY_VALUE, true);
				
				local depot = null;
				for (local airport = airports.Begin(); !airports.IsEnd(); airport = airports.Next()){
					depot = AIAirport.GetHangarOfAirport(AIStation.GetLocation(airport));
					if (depot != null && AIAirport.IsHangarTile(depot)){
						break;
					}
				}
				if (depot != null && AIAirport.IsHangarTile(depot)){
					local newplane = AIVehicle.BuildVehicle(depot, en);
					if (newplane != null && AIVehicle.IsValidVehicle(newplane)){
						local name = AIVehicle.GetName(plane);
						AIOrder.ShareOrders(newplane, plane);
						AIVehicle.StartStopVehicle(newplane);
						
						AIVehicle.SendVehicleToDepot(plane);
						local i=0;
						while (!AIVehicle.SetName(plane, "delete"+i)){
							i++;
						}
						
						AIVehicle.SetName(newplane, name);
						AILog.Info("Upgraded \""+name+"\" from \""+AIEngine.GetName(AIVehicle.GetEngineType(plane))+"\" to \""+AIEngine.GetName(en)+"\"");
						break;
					}
				}
			}
		}
	}
}

function AirManager::BuildNewAirport() {
	local towns = AITownList();
	foreach (town in townBlacklist){
		towns.RemoveItem(town);
	}
	//towns.Valuate(AITown.GetAllowedNoise);
	//towns.KeepAboveValue(AIAirport.GetNoiseLevelIncrease(AIAirport.AT_INTERCON));
	local stations = AIStationList(AIStation.STATION_AIRPORT);
	//stations.Valuate(AIStation.HasStationType, AIStation.STATION_AIRPORT);
	//stations.KeepValue(1);
	stations.Valuate(AIStation.GetNearestTown);
	foreach (station, town in stations){
		towns.RemoveItem(town);
		towns.Valuate(AITown.GetDistanceManhattanToTile, AIStation.GetLocation(station));
		towns.KeepAboveValue(40);
	}
	towns.Valuate(Util.GetRandomizedPopulation);
	towns.KeepAboveValue(1000);
	towns.Sort(AIList.SORT_BY_VALUE, false);
	
	local totalStations = stations.Count();
	
	stations.Valuate(Util.StationWaitingCargo);
	stations.KeepAboveValue(100);
	
	local stations2=AIStationList(AIStation.STATION_AIRPORT);
	stations2.Valuate(GetNumberOfPlanes);
	stations2.KeepValue(0);
	stations2.Valuate(Util.ConstValue, 1000);
	stations.AddList(stations2);
	
	if (totalStations > 0 && (stations.Count()<2||stations.Count()<totalStations/3)){
		AILog.Info("not building, valid stations: "+stations.Count()+", total stations: "+totalStations);
		return false;
	}
	
	local biggestTown = towns.Begin();
	for (local biggestTown = towns.Begin(); !towns.IsEnd(); biggestTown=towns.Next()){
		AILog.Info("Building airport in "+AITown.GetName(biggestTown));
		local air1 = BuildAirport(biggestTown);
		if (air1 != null && AIMap.IsValidTile(air1)){
			if (stations.Count()==0){
				local towns2 = AITownList();
				towns2.KeepList(towns);
				towns2.RemoveItem(biggestTown);
				towns2.Valuate(AITown.GetDistanceManhattanToTile, air1);
				towns2.KeepAboveValue(60);
				towns2.Valuate(Util.GetRandomizedPopulation);
				towns2.Sort(AIList.SORT_BY_VALUE, false);
				if (towns2.IsEmpty()){
					AILog.Info("no towns found");
				}
				for (local secondTown = towns2.Begin(); !towns2.IsEnd(); secondTown = towns2.Next()){
					AILog.Info("attempting secondary airport in "+AITown.GetName(secondTown));
					local air2 = BuildAirport(secondTown);
					if (air2!=null && AIMap.IsValidTile(air2)){
						AILog.Info("Connecting "+AITown.GetName(biggestTown)+" to "+AITown.GetName(secondTown));
						if (ConnectAirports(air1, air2)){
							return true;
						}
					}
				}
			
			}
			else {
				stations.Sort(AIList.SORT_BY_VALUE, false);
				stations.Valuate(AIStation.GetDistanceManhattanToTile, air1);
				stations.KeepAboveValue(30);
				
				stations.Valuate(SupportsExtraPlanes);
				stations.KeepValue(1);
				if (stations.IsEmpty()){
					local planes = AIVehicleList();
					planes.Valuate(AIVehicle.GetVehicleType);
					planes.KeepValue(AIVehicle.VT_AIR);
					
					if (planes.IsEmpty()){
						//No empty stations but also no planes, strange...
						AILog.Warning("unexpected code path (air.nut, #1)");
						continue;
					}
					local plane = planes.Begin();
					while (plane != null){
						if (AIOrder.IsGotoStationOrder(plane, 0)){
							local orderPos = AIOrder.IsGotoStationOrder(plane, 1)?1:2
							local distance = AIOrder.GetOrderDistance(AIVehicle.VT_AIR, AIOrder.GetStopLocation(plane, 0), air1);
							if (AIEngine.GetMaximumOrderDistance(AIVehicle.GetEngineType(plane))==0 ||
							    AIEngine.GetMaximumOrderDistance(AIVehicle.GetEngineType(plane))>distance){ //check the range
								AIOrder.RemoveOrder(plane, orderPos);
								AIOrder.AppendOrder(plane, air1, AIOrder.OF_NONE);
								AILog.Info("redirecting plane "+AIVehicle.GetName(plane));
								return true;
							}
						}
						if (plane != null){
							plane = planes.Next();
						}
					}
				}
				else {
					for (local biggest = stations.Begin(); !stations.IsEnd(); biggest=stations.Next()){
						if (ConnectAirports(air1, AIStation.GetLocation(biggest))){
							return true;
						}
					}
				}
				
			}
		}
	}
	return false;
}

function AirManager::UpgradeAirports(){
	local airports = AIStationList(AIStation.STATION_AIRPORT);
	airports.Valuate(SupportsExtraPlanes);
	airports.KeepValue(0); //only full airports
	airports.Valuate(AIStation.GetCargoWaiting, cargo_passengers);
	airports.KeepAboveValue(250); //Low value for testing, ideally should probably be 500-750
	foreach (port in airportUpgradeBlacklist){
		airports.RemoveItem(port);
	}
	
	if (airports.Count() > 0){
		return UpgradeAirport(airports.Begin())
	}
}

function AirManager::ConnectAirports(air1, air2){
	AILog.Info("connecting "+AIBaseStation.GetName(AIStation.GetStationID(air1))+" to "+AIBaseStation.GetName(AIStation.GetStationID(air2)));
	local engines = AIEngineList(AIVehicle.VT_AIR);
	foreach (airportType in helicopterAirportTypes){
		if (AIAirport.GetAirportType(air1)==airportType ||
		    AIAirport.GetAirportType(air2)==airportType){
			engines.Valuate(AIEngine.GetPlaneType);
			engines.KeepValue(AIAirport.PT_HELICOPTER);
			break;
		}
	}
	foreach (airportType in smallAirportTypes){
		if (AIAirport.GetAirportType(air1)==airportType ||
		    AIAirport.GetAirportType(air2)==airportType){
			engines.Valuate(AIEngine.GetPlaneType);
			engines.RemoveValue(AIAirport.PT_BIG_PLANE);
			break;
		}
	}
	engines.Valuate(AIEngine.GetMaximumOrderDistance);
	engines.RemoveBetweenValue(5, AIOrder.GetOrderDistance(AIVehicle.VT_AIR, air1, air2));
	engines.Valuate(AIEngine.GetCapacity);
	engines.Sort(AIList.SORT_BY_VALUE, false);
	if (engines.IsEmpty()){
		AILog.Warning("no engine found");
		return false;
	}
	AILog.Info("number of engines: "+engines.Count());
	
	local depot = AIAirport.GetHangarOfAirport(air1);
	if (depot == null){
		depot = AIAirport.GetHangarOfAirport(air2);
	}
	
	local vehicle = null;
	local engine = engines.Begin();
	while (vehicle==null && !engines.IsEnd()){
		if (Money.ReserveMoney(AIEngine.GetPrice(engine))){
			AILog.Info("Building engine: "+AIEngine.GetName(engine));
			vehicle = AIVehicle.BuildVehicle(depot, engine);
			if (!AIVehicle.IsValidVehicle(vehicle)){
				AILog.Warning("failed to build vehicle: "+AIError.GetLastErrorString());
			}
		}
		engine = engines.Next();
	}
	
	if (vehicle!=null){
		AIOrder.AppendOrder(vehicle, air1, AIOrder.OF_NONE);
		AIOrder.AppendOrder(vehicle, air2, AIOrder.OF_NONE);
		
		AIVehicle.StartStopVehicle(vehicle);
		AIVehicle.SetName(vehicle, Util.CapLength(AIBaseStation.GetName(AIStation.GetStationID(air1)),8)+" to "+
		                           Util.CapLength(AIBaseStation.GetName(AIStation.GetStationID(air2)),8));
		
		AILog.Info("end");
		return true;
	}
}

function AirManager::BuildAirport(townID){
	foreach (a in airportTypes){
		if (AIAirport.IsValidAirportType(a) && Money.ReserveMoney(AIAirport.GetPrice(a))){
			local tile=BuildAirportType(townID, a, null)
			if (tile != null && AIMap.IsValidTile(tile)){
				return tile;
			}
		}
	}
	townBlacklist.push(townID);
	return null;
}

function AirManager::UpgradeAirport(airportID){
	foreach (type in airportTypes){
		if (type == AIAirport.GetAirportType(AIStation.GetLocation(airportID))){
			if (AIStation.IsAirportClosed(airportID)){
				AIStation.OpenCloseAirport(airportID);
			}
			return false;
		}
		
		else if (AIAirport.IsValidAirportType(type) && Money.ReserveMoney(AIAirport.GetPrice(1)+400)){
			
		
			if (!AIStation.IsAirportClosed(airportID)){
				AIStation.OpenCloseAirport(airportID);
				AIController.Sleep(10);
			}
			
			local newAirport = BuildAirportType(AITile.GetClosestTown(AIStation.GetLocation(airportID)), type, airportID);
			if (newAirport!=null && AIMap.IsValidTile(newAirport) && AIAirport.IsAirportTile(newAirport)){
				if (AIStation.GetStationID(newAirport)==airportID){
					if (AIStation.IsAirportClosed(airportID)){
						AILog.Info("airport closed, opening");
						AIStation.OpenCloseAirport(airportID);
					}
					return true;
					AILog.Info("upgraded airport "+AIStation.GetName(airportID));
				}
				else {
					local planes = AIVehicleList_Station(airportID);
					planes.Valuate(AIVehicle.GetVehicleType);
					planes.KeepValue(AIVehicle.VT_AIR);
					foreach (plane, _ in planes){
						for (local i=0; AIOrder.IsValidVehicleOrder(plane, i); i++){
							if (AIStation.GetOrderStationID(AIOrder.GetDestination(plane, 1))==airportID){
								AIOrder.RemoveOrder(plane, i);
								AIOrder.InsertOrder(plane, i, newAirport, AIOrder.OF_NONE);
								//Don't break in case multiple orders are to this station
							}
						}
					}
					local name = AIStation.GetName(airportID);
					while (!AITile.DemolishTile(AIStation.GetLocation(airportID)) &&
						AIError.GetLastError() == AIError.ERR_VEHICLE_IN_THE_WAY){
						AIController.Sleep(10);
					}
					AIStation.SetName(airportID, "obsolete airport");
					AIStation.SetName(AIStation.GetStationID(newAirport), name);
					
					if (AIStation.IsAirportClosed(airportID)){
						AILog.Info("airport closed, opening");
						AIStation.OpenCloseAirport(airportID);
					}
					AILog.Info("upgraded airport "+AIStation.GetName(airportID));
					
					return true;
				}
			}
			else {
				AILog.Warning("airport upgrade of "+AIStation.GetName(airportID)+" failed");
				airportUpgradeBlacklist.push(airportID);
			}
		}
	}
	if (AIStation.IsAirportClosed(airportID)){
		AIStation.OpenCloseAirport(airportID);
	}
}



function AirManager::BuildAirportType(townID, airportType, stationID){
	AILog.Info("Trying to build "+airportType);
	local location = AITown.GetLocation(townID);
	//AISign.BuildSign(location, "town");
	
	local width = AIAirport.GetAirportWidth(airportType);
	local height = AIAirport.GetAirportHeight(airportType);
	
	local tiles = Util.GetTileListRadius(location - AIMap.GetTileIndex(width/2, height/2), 20);
	tiles.Valuate(AITile.GetClosestTown);
	tiles.KeepValue(townID);
	tiles.Valuate(IsClearOrAirport, stationID);;
	tiles.KeepValue(1);
	
	local tilesBefore = tiles.Count();
	
	local cargos = AICargoList();
	cargos.Valuate(AICargo.HasCargoClass, AICargo.CC_PASSENGERS);
	cargos.KeepValue(1);
	
	foreach (cargo, value in cargos){
		if (AICargo.IsValidCargo(cargo)){
			tiles.Valuate(AITile.GetCargoAcceptance, cargo, width, height, AIAirport.GetAirportCoverageRadius(airportType));
			tiles.KeepAboveValue(100); //reject tiles with little acceptance
			//typical range seems to be 50-200 ish
		}
	}
	
	tiles.Valuate(AITile.GetCargoAcceptance, cargos.Begin(), width, height, AIAirport.GetAirportCoverageRadius(airportType));
	tiles.Sort(AIList.SORT_BY_VALUE, false);
	AILog.Info("number of tiles before: "+tilesBefore+" after: "+tiles.Count()+
				" closest: "+AITile.GetDistanceManhattanToTile(tiles.Begin(), location - AIMap.GetTileIndex(width / 2, height / 2)));
	//if (!tiles.IsEmpty()){
	//	AISign.BuildSign(tiles.Begin(), "first");
	//}
	foreach (tile, value in tiles){
	    local tiles2=AITileList();
		tiles2.AddRectangle(tile, tile + AIMap.GetTileIndex(width - 1, height - 1));
		tiles2.Valuate(IsClearOrAirport, stationID);
		tiles2.KeepValue(1);
		
		if (tiles2.Count()==(width*height)){
			foreach (tile, _ in tiles2){
				if (!AITile.IsBuildable(tile)){
					AILog.Warning("attempting to demolish "+tile);
					while (!AITile.DemolishTile(tile)){
						AIController.Sleep(10);
						if (AIError.GetLastError()!=AIError.ERR_VEHICLE_IN_THE_WAY){
							AILog.Warning("clearing tile failed: "+AIError.GetLastErrorString());
						}
					}
				}
			}
			if (AIAirport.BuildAirport(tile, airportType, stationID!=null?stationID:AIStation.STATION_NEW) ||
				(AIError.GetLastError()==AIError.ERR_STATION_TOO_SPREAD_OUT && AIAirport.BuildAirport(tile, airportType,
				AIStation.STATION_NEW))){
				AIBaseStation.SetName(AIStation.GetStationID(tile), AITown.GetName(townID));
				//AISign.BuildSign(tile, "topleft");
				AILog.Info("Airport built");
				return tile;
			}
			else if (AIError.GetLastError() != AIError.ERR_NONE) {
				AISign.BuildSign(tile, AIError.GetLastErrorString());
				AISign.BuildSign(tile + AIMap.GetTileIndex(width, height), "invalid bottom");
				return;
				AILog.Warning("failed to build airport: "+AIError.GetLastErrorString());
			}
		}
		else {
			//AISign.BuildSign(tile, number1+", "+tiles2.Count()+" ("
			//	+AITile.GetCargoAcceptance(tile, cargos.Begin(), width, height, AIAirport.GetAirportCoverageRadius(airportType))+")");
		}
	}
	AILog.Info("failed");
	return null;
}